class SchemazerConfig:
    QUERY_TOKEN = 'token'
    HEADER_TOKEN = 'Authentication-Token'
    COOKIE_TOKEN = 'auth_token'

    QUERY_APIKEY = 'apikey'
    HEADER_APIKEY = 'Apikey'
    COOKIE_APIKEY = 'apikey'

    ERROR_HTTP_CODE = 200
    HOST = 'localhost'
    HTTP_SCHEMA = 'http'
    VERSION = 'schemazer'

    PROJECT_NAME = 'Unknown'
    DOCS_PATH = 'documentation'



ERROR_WRONG = 'SchemazerParameter `{0}` contains non-valid characters.'
ERROR_TYPE = 'SchemazerParameter `{0}` have invalid type.'
ERROR_MISSING = 'Required parameter `{0}` missing.'
ERROR_INTERVAL = 'SchemazerParameter `{0}` out of available values.'



from flask import Blueprint, abort, current_app, json, render_template

doc_bp = Blueprint(
    'doc', __name__,
    template_folder='./templates',
    static_folder='static',
    static_url_path='/static/doc'
)

base_params = {
    'project_name': current_app.schemazer.config.PROJECT_NAME,
    'docs_path': current_app.schemazer.config.DOCS_PATH
}


def _schema_groups():
    return [x for x in dir(current_app.schemazer.schema)
            if not x.startswith('_')]


def _errors():
    groups = current_app.schemazer.schema.__groups__
    errors = set()

    for group in groups:
        for method in group().__methods__:
            errors.update(set(method.errors))

    return errors


@doc_bp.route(f'/{current_app.schemazer.config.DOCS_PATH}/')
def documentation():
    groups = current_app.schemazer.schema.__groups__
    groups.sort()

    params = {
        'version': current_app.schemazer.config.VERSION,
        'groups': groups,
    }
    params.update(base_params)

    return render_template('pages/index.html', **params)


@doc_bp.route(f'/{current_app.schemazer.config.DOCS_PATH}/methods/',
              methods=['GET'])
def method_groups():
    groups = [x for x in dir(current_app.schemazer.schema)
              if not x.startswith('_')]
    groups.sort()

    params = {
        'title': 'Methods api',
        'host_ip': current_app.schemazer.config.HOST,
        'groups': groups,
    }
    params.update(base_params)

    return render_template('pages/groups.html', **params)


@doc_bp.route(
    f'/{current_app.schemazer.config.DOCS_PATH}/methods/<string:group>/',
    methods=['GET'])
def method_for_group(group):
    group_obj = getattr(current_app.schemazer.schema, group, None)

    if not group_obj:
        abort(404)

    methods = [x for x in dir(group_obj) if not x.startswith('_')]
    methods.sort()

    params = {
        'title': 'SchemazerMethod group %s' % group,
        'host_ip': current_app.schemazer.config.HOST,
        'group': group,
        'methods': methods,
    }
    params.update(base_params)

    return render_template('pages/methods.html', **params)


@doc_bp.route(
    f'/{current_app.schemazer.config.DOCS_PATH}/methods/'
    f'<string:group>.<string:method>/',
    methods=['GET'])
def method_route(group, method):
    group_obj = getattr(current_app.schemazer.schema, group, None)

    if not group_obj or not getattr(group_obj, method, None):
        abort(404)

    method_schema = getattr(group_obj, method, None)

    response = {
        'schema': json.dumps(
            method_schema.response.to_dict(),
            indent=2,
            sort_keys=True,
            ensure_ascii=False)
    }

    if method_schema.response:
        response['example'] = json.dumps(
            method_schema.response.example,
            indent=2,
            sort_keys=True,
            ensure_ascii=False
        )

    params = {
        'title': f'{group}.{method_schema.name}',
        'host_ip': current_app.schemazer.config.HOST,
        'group': group,
        'method': method,
        'method_schema': method_schema,
        'response': response,
        'errors': [x.to_dict() for x in method_schema.errors],
    }
    params.update(base_params)

    return render_template('pages/method.html', **params)


@doc_bp.route(f'/{current_app.schemazer.config.DOCS_PATH}/errors/',
              methods=['GET'])
def errors_route():
    groups = current_app.schemazer.schema.__groups__
    errors = set()

    for group in groups:
        for method in group().__methods__:
            errors.update(set(method.errors))

    errors_repr = [error.to_dict() for error in errors]
    errors_list = sorted(errors_repr, key=lambda k: k['code'])

    params = {
        'title': 'Errors',
        'errors': errors_list or [],
    }
    params.update(base_params)

    return render_template('pages/errors.html', **params)


@doc_bp.route(
    f'/{current_app.schemazer.config.DOCS_PATH}/errors/'
    f'<string:error_group>.<string:error_name>/',
    methods=['GET'])
def error_route(error_group, error_name):
    find_error = None
    for error in _errors():
        if error.code == f'{error_group}.{error_name}':
            find_error = error
            break

    if not find_error:
        abort(404)

    params = {
        'title': 'Ошибка %s' % find_error.code,
        'error': find_error,
    }
    params.update(base_params)

    return render_template('pages/error.html', **params)



from setuptools import find_packages, setup

setup(
    name='schemazer',
    version='1.1.6',
    packages=find_packages(),
    author='Dmitriy Danshin',
    author_email='Tingerlink@yandex.ru',
    url='https://gitlab.tingerlink.pro/tingerlink/schemazer',
    include_package_data=True,
    install_requires=[
        'Flask>=0.12.0',
        'werkzeug'
    ]
)



import re

from typing import List
from flask import request

from schemazer.base import SchemazerParameter
from schemazer.messages import (ERROR_WRONG, ERROR_INTERVAL, ERROR_TYPE,
                                ERROR_MISSING)
from .commons.errors import RequestErrors


class RequestValidator:
    """
    The class implements methods for verifying the query parameters for API
    """

    @staticmethod
    def validate_required(parameter: SchemazerParameter) -> bool:
        """
        Check in request required parameter
        :param SchemazerParameter parameter: Schema parameter
        """
        if not request.query.args.get(parameter.name) and parameter.required:
            return RequestErrors.BadRequest.json_abort(
                failure_field=parameter.name,
                extend_msg=ERROR_MISSING.format(parameter.name))

        return True

    def validate_parameters(
            self, parameters: List[SchemazerParameter]) -> bool:
        """
        Validate request parameters by schema.
        Check required parameters and parameter format.
        :param List[SchemazerParameter] parameters: Schema parameters
        :return bool: True if all parameters is correct, or abort response.
        """
        for parameter in parameters or []:
            result = self.validate_required(parameter)
            if result is not True:
                return result

            result = parameter.validator.check(parameter)
            if result is not True:
                return result

        return True

    @staticmethod
    def validate_interval(parameters: List[SchemazerParameter]) -> bool:
        """
        Validate request parameters by available values interval
        :param List[SchemazerParameter] parameters: Schema parameters
        """
        for parameter in parameters or list():
            value = request.query.args.get(parameter.name)
            if callable(parameter.interval) and not parameter.interval(value):
                return RequestErrors.BadRequest.json_abort(
                    failure_field=parameter.name,
                    extend_msg=ERROR_INTERVAL.format(parameter.name))
        return True

    @staticmethod
    def set_default_values(parameters: List[SchemazerParameter]):
        for parameter in parameters:
            value = request.query.args.get(parameter.name)
            value = str(value) if value is not None else None

            try:
                if value or parameter.default:
                    request.query.args[parameter.name] = \
                        parameter.type(value) or parameter.default
            except ValueError:
                return RequestErrors.BadRequest.json_abort(
                    failure_field=parameter.name,
                    extend_msg=ERROR_TYPE.format(parameter.name))

    def process(self, parameters: List[SchemazerParameter]) -> bool:
        """
        Validate request parameters by schema.
        - validate required parameters
        - validate format pattern of parameters
        - validate parameters interval
        - validate params types, convert from string query params by schema
          parameter types.
        :param parameters: Schema parameters
        """
        for func in [self.validate_parameters, self.validate_interval,
                     self.set_default_values]:
            result = func(parameters)
            if result is not bool:
                return result

        return True


class Validator:
    """
    Base validator representation.
    """
    def check(self, **kwargs):
        raise NotImplementedError()


class PatternValidator(Validator):
    """
    Regex validator.
    """
    def __init__(self, pattern: str):
        """
        Init by regex pattern.
        :param str pattern:
        """
        self.pattern = pattern

    @staticmethod
    def _regular_check(value, expression):
        process = re.compile(expression)

        if type(value) in [bool, int, float]:
            value = str(value).lower()
        result = process.match(value)

        return bool(result)

    def check(self, parameter: SchemazerParameter) -> bool:
        """
        Check request parameter by pattern.
        :param SchemazerParameter parameter: SchemazerParameter schema
        """
        str_value = str(request.query.args.get(parameter.name))

        if not self._regular_check(str_value, self.pattern):
            return RequestErrors.BadRequest.json_abort(
                failure_field=parameter.name,
                extend_msg=ERROR_WRONG.format(parameter.name))

        return True



from functools import wraps
from flask import abort, request
from flask_classy import route

from schemazer.helpers import remote_addr
from schemazer.commons.errors import RequestErrors
from schemazer.config import SchemazerConfig
from schemazer.validator import RequestValidator


class QueryMeta:
    ip = None
    token = None
    cookies = None
    headers = None
    apikey = None


class Query:
    def __init__(self):
        self.meta = QueryMeta()
        self.args = {}

    def to_dict(self):
        return (f'==========request===========\n'
                f'ip: {self.meta.ip}\n'
                f'token: {self.meta.token}\n'
                f'cookies: {self.meta.cookies}\n'
                f'headers: {self.meta.headers}\n'
                f'args: {self.args}\n')


class Schemazer:
    def __init__(self, app, schema):
        self.schema = schema
        self.config = SchemazerConfig()
        app.schemazer = self

        for param in [x for x in dir(SchemazerConfig) if
                      not x.startswith('_')]:
            name = 'SCHEMAZER_' + param
            setattr(self.config, param,
                    app.config.get(name, getattr(self.config, param)))

    def route(self, schema=None, path=None):
        def decorator(f):
            @route(path if path else schema.path, methods=schema.methods)
            @wraps(f)
            def decorated_function(*args, **kwargs):
                """
                Create Query object from flask request.
                :return Query: Request with auth data.
                """
                request.query = Query()

                if request.method == 'GET':
                    request.query.args.update(request.args.to_dict(flat=True))
                if request.method == 'POST':
                    data = request.get_json() or dict()
                    if not isinstance(data, dict):
                        abort(RequestErrors.BadRequest.json_abort())
                    request.query.args.update(data)

                request.query.meta.token = (
                        request.query.args.get(self.config.QUERY_TOKEN) or
                        request.headers.get(self.config.HEADER_TOKEN) or
                        request.cookies.get(self.config.COOKIE_TOKEN)
                )
                request.query.meta.apikey = (
                        request.query.args.get(self.config.QUERY_APIKEY) or
                        request.headers.get(self.config.HEADER_APIKEY) or
                        request.cookies.get(self.config.COOKIE_APIKEY)
                )
                request.query.meta.ip = remote_addr(request)
                request.query.meta.cookies = request.cookies
                request.query.meta.headers = request.headers

                request_validator = RequestValidator()
                result = request_validator.process(schema.parameters)
                if result is not True:
                    return result

                return f(*args, **kwargs)
            return decorated_function
        return decorator



from typing import List
from flask import abort, current_app, jsonify, make_response, request

"""
Class map:

SchemazerSchema
    SchemazerGroup
        SchemazerMethod
            SchemazerParameter
            SchemazerLimit
            SchemazerErrorGroup
                SchemazerErrorBase
            SchemazerResponse
                SchemazerResponseField
                    SchemazerResponseObjectField
"""


class SchemazerMethod(object):
    name = None
    group_name = None

    description: None
    parameters = None
    limits = None
    errors: list = None
    response = None
    methods = None

    @property
    def url(self):
        uri = f'{current_app.schemazer.config.HOST}{self.path}'

        if current_app.schemazer.config.HTTP_SCHEMA:
            uri = f'{current_app.schemazer.config.HTTP_SCHEMA}://{uri}'

        return uri

    @property
    def path(self):
        current_app.logger.info(f'/{self.group_name}.{self.name}')
        return f'/{self.group_name}.{self.name}'

    def __init__(self, **kwargs):
        self.methods = ['GET']
        self.__dict__.update(kwargs)


class SchemazerGroup(object):
    """
    SchemazerGroup of methods.
    """

    @property
    def __methods__(self) -> List[SchemazerMethod]:
        return [getattr(self, x) for x in dir(self) if not x.startswith('_')]


class SchemazerSchema:
    @property
    def __groups__(self) -> List[SchemazerGroup]:
        return [getattr(self, x) for x in dir(self) if not x.startswith('_')]


class SchemazerParameter(object):
    """
    SchemazerParameter of query of schema method
    """
    # parameter query name
    name = None
    description = None
    # type of parameter value
    type = None
    # parameter is required
    required = None
    # example parameter for documentation
    example = None
    # pattern for validate parameter value
    validator = None
    # default value set if parameter not in query
    default = None
    # func for check parameter value by available value interval
    interval = True

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def to_dict(self):
        return {
            'name': self.name,
            'description': self.description,
            'type': str(self.type.__name__),
            'required': self.required,
            'example': self.example,
            'validator': self.validator
        }


class SchemazerLimit(object):
    """
    Limits for access by user roles.
    """
    role = None

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class SchemazerErrorGroup:
    """
    Set attributes as SchemazerErrorBase objects. Use SchemazerErrorGroup for
    access to error
    object.
    """
    @property
    def __all__(self):
        return [getattr(self, x) for x in dir(self) if
                isinstance(getattr(self, x), SchemazerErrorBase)]


class SchemazerErrorBase(object):
    """
    Class for create error fields in SchemazerErrorGroup.
    """
    name = None
    # SchemazerErrorGroup name
    group_name = None
    description = None

    def __init__(self, group_name, name, description=None):
        self.group_name = group_name
        self.name = name
        self.description = description

    def to_dict(self) -> dict:
        return {
            'code': self.code,
            'description': self.description
        }

    @property
    def code(self) -> str:
        return f'{self.group_name}.{self.name}'

    def json_abort(self, failure_field: str=None, extend_msg: str=''):
        """
        Abort response with error object.
        :param dict failure_field: SchemazerParameter in which the error
        occurred
        :param str extend_msg: Extend message with error description.
        :return:
        """
        data = {
            'error':
                {
                    'code': self.code,
                    'msg': f'{self.description} {extend_msg}' if extend_msg
                    else self.description,
                    'failure_param': failure_field,
                    'params': request.query.args or {},
                }
        }

        return make_response(jsonify(data),
                             current_app.config.get('ERROR_HTTP_CODE'))


class SchemazerResponseField(object):
    """
    Class use for create documentation. Simple field with
    single field. For create object with many params or  list or objects use
    `SchemazerResponseObjectField`.
    """
    name = None
    type = None
    description = None
    data_example = None  # example of return data

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def to_dict(self) -> dict:
        return {
            self.name: {
                'type': self.type.__name__ or 'undefined',
                'description': self.description,
            }
        }

    @property
    def example(self):
        return {self.name: self.data_example}


class SchemazerResponseObjectField(object):
    """
    Class use for create documentation .
    """
    name = None
    fields = None
    is_list = False

    def __init__(self, name=None, fields: list=None, is_list=False):
        self.name = name
        self.fields = fields or list()
        self.is_list = is_list

    def to_dict(self):
        obj = {}
        for x in self.fields:
            obj.update(x.to_dict())

        if self.name:
            return {obj if not self.is_list else [obj]}
        else:
            return obj if not self.is_list else [obj]

    @property
    def example(self):
        example = {}
        for x in self.fields:
            example.update(x.example)

        if self.name:
            return {example if not self.is_list else [example]}
        else:
            return example if not self.is_list else [example]


class SchemazerResponse(object):
    schema = None

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def to_dict(self):
        return {'response': self.schema.to_dict()}

    @property
    def example(self):
        return {'response': self.schema.example}
